import copy
import json
import os
from pathlib import Path

import graph

from pgtg.constants import OBSTACLE_NAMES, TILE_HEIGHT, TILE_WIDTH
from pgtg.map_generator import MapPlan
from pgtg.map_tiles_data import OBSTACLE_MASKS, TILES, TRAFFIC_LANES


def parse_map_object(
    map_plan: MapPlan,
) -> tuple[int, int, list[list[set[str]]], int, dict[tuple[int, int], str]]:
    """Parses a map plan, to the important features of a episode map.

    Args:
        map_plan: The map to be parsed

    Returns:
        A tuple (width, height, map, num_subgoals, subgoal_directions), where width and height are the dimensions of the map,
        map represents the features of each tile, num_subgoals is the number of subgoals in the map, and subgoal_directions
        is a dictionary mapping tile coordinates to the direction of the subgoal on that tile.
    """

    graph = parse_tile_map_to_graph(map_plan)

    # finds the shortest path between the lower left tile and the upper right tile.
    shortest_path: list[tuple[int, int]] = graph.shortest_path(
        (0, map_plan.height - 1), (map_plan.width - 1, 0)
    )[1]

    # dict maps tile coordinate to the character of the exit the subgoal has to be on.
    subgoal_coordinates_to_direction: dict[tuple[int, int], str] = {}
    for i in range(len(shortest_path) - 1):
        subgoal_coordinates_to_direction.update(
            {shortest_path[i]: find_direction(shortest_path[i], shortest_path[i + 1])}
        )

    # create a empty map
    map: list[list[set[str]]] = [
        [set() for _ in range(map_plan.height * TILE_HEIGHT)]
        for _ in range(map_plan.width * TILE_WIDTH)
    ]

    for tile_x in range(map_plan.width):
        for tile_y in range(map_plan.height):

            # copy the tile to preserve the original
            current_tile = copy.deepcopy(
                TILES[tuple(map_plan.tiles[tile_y][tile_x]["exits"])]
            )

            # if the tile is on the shortest path replace exit markers with subgoals
            if (tile_x, tile_y) in shortest_path[:-1]:
                replace_features_in_tile(
                    current_tile,
                    "exit " + subgoal_coordinates_to_direction[(tile_x, tile_y)],
                    "subgoal",
                )

            # if the tile is in the bottom left corner replace exit markers with start
            if tile_x == 0 and tile_y == map_plan.height - 1:
                replace_features_in_tile(
                    current_tile,
                    "exit west",
                    "start",
                )

            # if the tile is in the top right corner replace exit markers with final goals
            if tile_x == map_plan.width - 1 and tile_y == 0:
                replace_features_in_tile(
                    current_tile,
                    "exit east",
                    "final goal",
                )

            # remove unused exit markers
            replace_features_in_tile(
                current_tile,
                "exit north",
                None,
            )
            replace_features_in_tile(
                current_tile,
                "exit east",
                None,
            )
            replace_features_in_tile(
                current_tile,
                "exit south",
                None,
            )
            replace_features_in_tile(
                current_tile,
                "exit west",
                None,
            )

            # if the tile has a obstacle type add obstacles
            if map_plan.tiles[tile_y][tile_x].get("obstacle_type") is not None:
                assert (
                    map_plan.tiles[tile_y][tile_x].get("obstacle_mask") is not None
                ), f"The tile at ({tile_x},{tile_y}) has a obstacle type without a obstacle mask"

                add_obstacles_to_tile(
                    current_tile,
                    OBSTACLE_MASKS[map_plan.tiles[tile_y][tile_x]["obstacle_mask"]],
                    map_plan.tiles[tile_y][tile_x]["obstacle_type"],
                )

            # if the tile has any exits add traffic lanes
            if map_plan.tiles[tile_y][tile_x]["exits"] != [0, 0, 0, 0]:
                add_traffic_lanes_to_tile(
                    current_tile,
                    TRAFFIC_LANES[tuple(map_plan.tiles[tile_y][tile_x]["exits"])],
                )

            # if the tile is on the border of the map add car spawners
            if tile_x == 0:
                replace_features_in_tile(
                    current_tile,
                    "car_lane all right",
                    "car_spawner",
                    keep_old_features=True,
                )
            if tile_x == map_plan.width - 1:
                replace_features_in_tile(
                    current_tile,
                    "car_lane all left",
                    "car_spawner",
                    keep_old_features=True,
                )
            if tile_y == 0:
                replace_features_in_tile(
                    current_tile,
                    "car_lane all down",
                    "car_spawner",
                    keep_old_features=True,
                )
            if tile_y == map_plan.height - 1:
                replace_features_in_tile(
                    current_tile,
                    "car_lane all up",
                    "car_spawner",
                    keep_old_features=True,
                )

            # add the tile to the map
            for square_x in range(TILE_WIDTH):
                for square_y in range(TILE_HEIGHT):
                    map[tile_x * TILE_WIDTH + square_x][
                        tile_y * TILE_HEIGHT + square_y
                    ] = current_tile[square_x][square_y]

    # add the direction of the final goal
    subgoal_coordinates_to_direction.update({shortest_path[-1]: "east"})

    return (
        map_plan.width * TILE_WIDTH,
        map_plan.height * TILE_HEIGHT,
        map,
        len(subgoal_coordinates_to_direction),
        subgoal_coordinates_to_direction,
    )


def replace_features_in_tile(
    tile: list[list[set[str]]],
    old_feature: str,
    new_feature: str | None,
    keep_old_features: bool = False,
) -> None:
    """Replaces a feature in a tile with up to one new feature.

    Args:
        tile: The tile to be modified.
        old_feature: The feature to be replaced.
        new_feature: The new feature to replace the old feature with. If None, the old feature simply is removed.
    """

    for x in range(TILE_WIDTH):
        for y in range(TILE_HEIGHT):
            if old_feature in tile[x][y]:
                if not keep_old_features:
                    tile[x][y].remove(old_feature)

                if new_feature is not None:
                    tile[x][y].add(new_feature)


def add_obstacles_to_tile(
    tile: list[list[set[str]]], obstacle_mask: list[list[set[str]]], obstacle_type: str
) -> None:
    """Adds obstacles to a tile.

    Args:
        tile: The tile to be modified.
        obstacle_mask: The mask of the obstacles to be added.
        obstacle_type: The type of the obstacles to be added.
    """

    assert obstacle_type in OBSTACLE_NAMES, f"Unknown obstacle type: {obstacle_type}"

    for x in range(TILE_WIDTH):
        for y in range(TILE_HEIGHT):
            if "obstacle" in obstacle_mask[x][y] and "wall" not in tile[x][y]:
                tile[x][y].add(obstacle_type)


def add_traffic_lanes_to_tile(
    tile: list[list[set[str]]], traffic_lanes: list[list[set[str]]]
) -> None:
    """Adds traffic lanes to a tile.

    Args:
        tile: The tile to be modified.
        traffic_lanes: The traffic lanes to be added.
    """

    for x in range(TILE_WIDTH):
        for y in range(TILE_HEIGHT):
            tile[x][y].update(traffic_lanes[x][y])


def json_file_to_map_plan(path: str) -> MapPlan:
    """Creates a map plan from a JSON file.

    Args:
        path: The path to the JSON file, either absolute or relative to the current working directory.

    Returns:
        A map plan.
    """

    if not path.endswith(".json"):
        path = path + ".json"

    map_object = json.load(open(path))
    return MapPlan.from_dict(map_object)


def parse_tile_map_to_graph(tile_map: MapPlan) -> graph.Graph:
    """Generates a graph from a map plan representing the connections in the map.

    Args:
        tile_map: A map plan of the map the graph is generated for.

    Returns:
        graph: A graph representing the connections in the map.
    """

    result = graph.Graph()
    width = tile_map.width
    height = tile_map.height
    i = 0
    for row in tile_map.tiles:
        j = 0
        for node in row:
            result.add_node((j, i), node)

            if node["exits"][0] and i > 0:
                result.add_edge((j, i), (j, i - 1))

            if node["exits"][1] and j < width - 1:
                result.add_edge((j, i), (j + 1, i))

            if node["exits"][2] and i < height - 1:
                result.add_edge((j, i), (j, i + 1))

            if node["exits"][3] and j > 0:
                result.add_edge((j, i), (j - 1, i))
            j = j + 1
        i = i + 1
    return result


def find_direction(coordinates: tuple[int, int], other: tuple[int, int]) -> str:
    """Compares two tile coordinates to find out the cardinal direction they lie in relative to each other.

    Args:
        coordinates: Coordinates of the first tile.
        other: Coordinates of the second tile.

    Returns:
        The name of the cardinal direction the second tile lies in relative to the first tile.

    Raises:
        ValueError: If the two tiles are not horizontal or vertical to each other.
    """

    coordinates_x, coordinates_y = coordinates
    other_x, other_y = other

    if coordinates_y == other_y:
        if coordinates_x - other_x < 0:
            return "east"
        elif coordinates_x - other_x > 0:
            return "west"
    if coordinates_x == other_x:
        if coordinates_y - other_y < 0:
            return "south"
        elif coordinates_y - other_y > 0:
            return "north"
    raise ValueError("Not a cardinal Direction.")



import logging
import statistics as stat
import time
from dataclasses import dataclass

import numpy as np


@dataclass
class EvaluationResult:
    wins: int
    losses: int
    episode_timed_out: int
    mean_return: float
    mean_steps: float
    evaluation_timed_out: bool


def _episode(env, agent, max_steps, gamma=0.99, timeout=float("inf")):
    """
    Run a single episode for at most max_steps

    Args:
        env: environment in which episode takes place
        agent: the acting agent
        max_steps (int): episode is cut off after this maximum number of steps have been carried out
        gamma (float): discounting factor when calculating return

    Returns:

    """
    done = False
    ret = 0

    t_start = time.time()
    for t in range(max_steps):
        action = agent.act_with_env(env.clone())
        r, _, done = env.step(action)
        ret += r * np.power(gamma, t)
        elapsed = time.time() - t_start
        if done or elapsed > timeout:
            break

    return ret, t, int(ret > 0), int(ret < 0), int(not done), elapsed


class Evaluator:
    def __init__(self, env):
        """Init the Evaluator class

        Args:
            env (env): environment to evaluate upon.
        """
        self.agents = []
        self.names = []
        self._results = []
        self._timed_out = []
        self._evaluated = []
        self.env = env

    def __len__(self):
        return len(self.agents)

    def add_agent(self, agent):
        """add an agent for evaluation. Agents must implement the `act_with_env` and `get_name` method.

        Args:
            agent (Agent): agent to add.
        """
        self.agents.append(agent)
        self.names.append(agent.get_name())
        self._timed_out.append(False)
        self._evaluated.append(False)

    def reset(self):
        """reset the evaluator for new evaluations."""
        self._results = [([], [0, 0, 0], []) for _ in range(len(self))]
        self._timed_out = [False for _ in range(len(self))]
        self._evaluated = [False for _ in range(len(self))]

    def evaluate(self, number, max_steps=100, GAMMA=0.99, verbose=False):
        """evaluate using number many episodes with at most max_steps many steps.
        Return will be discounted using gamma.

        Args:
            number (int): number of episodes to use for eval
            max_steps (int, optional): maximal number of steps per episode. Defaults to 100.
            GAMMA (float, optional): Discount factor for calculating the return. Defaults to 0.99.
            verbose (bool, optional): Control output

        Returns:
            array [float, (int, int, int), float]: array containing 1. the return received in average, 2. the win-lose-timeout numbers, 3. the average number of taken steps in the winning case.
        """
        self.reset()
        env = self.env
        divider = int(0.1 * number)

        for i in range(number):
            env.reset()

            for a, agent in enumerate(self.agents):
                current_env = env.clone()

                done = False
                ret = 0

                for t in range(max_steps):
                    action = agent.act_with_env(current_env.clone())

                    r, _, done = current_env.step(action)
                    ret += r * np.power(GAMMA, t)
                    if done:
                        break
                # entry 0 is the place for the returns
                self._results[a][0].append(ret)

                if r > 0:
                    # win counter
                    self._results[a][1][0] += 1
                    # include step counter only when winning
                    # entry 2 for t
                    self._results[a][2].append(t)
                else:
                    if r < 0:
                        # loose counter
                        self._results[a][1][1] += 1
                    else:
                        # time_out counter
                        self._results[a][1][2] += 1

            if verbose and i % divider == 0:
                print(i, "/", number)

        self._evaluated = [True for _ in range(len(self))]
        self.print()
        return self._results

    def results_as_dict(self):
        return {
            name: result
            for name, result, eval_done in zip(
                self.names, self._results, self._evaluated
            )
            if eval_done
        }

    def results(self):
        return {
            name: EvaluationResult(
                wins=res[1][0],
                losses=res[1][1],
                episode_timed_out=res[1][2],
                mean_return=np.mean(res[0]),
                mean_steps=np.mean(res[2]),
                evaluation_timed_out=timed_out,
            )
            for name, res, timed_out, eval_done in zip(
                self.names, self._results, self._timed_out, self._evaluated
            )
            if eval_done
        }

    def print(self):
        """display the results. Works only, if some eval was done."""
        print(self.format_all())

    @staticmethod
    def format(name, wins, losses, timeouts, mean_return, mean_steps):
        return (
            "Agent %s won %i, lost %i and timed out %i games, by receiving an average return of %.2f.\nIn the winning case, %.2f steps were taken on average"
            % (name, wins, losses, timeouts, mean_return, mean_steps)
        )

    def format_all(self):
        if self._results == []:
            return "No agents were evaluated"
        else:
            lines = []
            for _, name, res, eval_done in zip(
                self.agents, self.names, self._results, self._evaluated
            ):
                if eval_done:
                    lines.append(
                        Evaluator.format(
                            name,
                            res[1][0],
                            res[1][1],
                            res[1][2],
                            np.mean(res[0]),
                            np.mean(res[2]),
                        )
                    )
                else:
                    lines.append("Agent %s was not evaluated" % name)
            return "\n\n".join(lines)


class TimedEvaluator(Evaluator):
    """
    Evaluator with per-agent timeout
    """

    def __init__(self, env):
        super(TimedEvaluator, self).__init__(env)

    def add_agent(self, agent):
        super(TimedEvaluator, self).add_agent(agent)

    @staticmethod
    def _evaluate_agent(env, agent, number, max_steps, gamma, timeout):
        t_total = 0.0
        result = [[], [0, 0, 0], []]
        timed_out = False
        for _ in range(number):
            if t_total >= timeout:
                logging.warning(
                    "Agent %s timed out after %.4f seconds."
                    % (agent.get_name(), t_total)
                )
                timed_out = True
                break

            env.reset()
            ret, steps, win, loose, ep_timeout, elapsed = _episode(
                env, agent, max_steps, gamma, timeout - t_total
            )
            t_total += elapsed

            result[0].append(ret)
            result[1][0] += win
            result[1][1] += loose
            result[1][2] += ep_timeout
            if win == 1:
                result[2].append(steps)

        if not timed_out:
            logging.info(
                "Finished evaluation of '%s' within %.4f seconds."
                % (agent.get_name(), t_total)
            )

        return result, timed_out

    def evaluate(
        self, number, max_steps=100, GAMMA=0.99, verbose=False, timeout=float("inf")
    ):
        """evaluate using number many episodes with at most max_steps many steps.
        Return will be discounted using gamma.

        Args:
            number (int): number of episodes to use for eval
            max_steps (int, optional): maximal number of steps per episode. Defaults to 100.
            GAMMA (float, optional): Discount factor for calculating the return. Defaults to 0.99.
            verbose (bool, optional): Control output
            timeout (float, optional): Per-agent timeout in minutes. If this limit is exceeded, the

        Returns:
            array [float, (int, int, int), float]: array containing 1. the return received in average, 2. the win-lose-timeout numbers, 3. the average number of taken steps in the winning case.
        """
        self.reset()

        timeout = timeout * 60

        for (
            a,
            agent,
        ) in enumerate(self.agents):
            try:
                result, timed_out = TimedEvaluator._evaluate_agent(
                    self.env.clone(), agent, number, max_steps, GAMMA, timeout
                )
                self._results[a] = result
                self._timed_out[a] = timed_out
                self._evaluated[a] = True
            except:
                logging.exception(
                    "An exception occurred while evaluating agent '%s'"
                    % agent.get_name()
                )

        return self._results


class ModularEvaluator:
    def __init__(self, env, agent):
        self.env = env
        self.agent = agent

    def reset(self):
        self.env.reset()
        self._results = ([], [0, 0, 0, 0])

    def evaluate(self, number, max_steps=100, GAMMA=0.99):
        """evaluate using number many episodes with at most max_steps many steps.
        Return will be discounted using gamma.
        Use print() function to visualize return.

        Args:
            number (int): number of episodes to use for eval
            max_steps (int, optional): maximal number of steps per episode. Defaults to 100.
            GAMMA (float, optional): Discount factor for calculating the return. Defaults to 0.99.


        Returns:
            array [float, (int, int, int, int)]: [average reward, (terminated, truncated, over max steps, negative reward)]
        """
        self.reset()
        env = self.env
        agent = self.agent

        for i in range(number):
            state, _ = env.reset()

            total_reward = 0

            for t in range(max_steps):
                action = agent.act(state)
                state, reward, terminated, truncated, _ = env.step(action)

                total_reward += reward * np.power(GAMMA, t)

                if terminated:
                    # loose counter
                    self._results[1][0] += 1
                    break
                if truncated:
                    # time_out counter
                    self._results[1][1] += 1
                    break
                if t == max_steps - 1:
                    self._results[1][2] += 1
                    break

            if total_reward < 0:
                self._results[1][3] += 1

            self._results[0].append(total_reward)

        return self._results

    def print(self):
        avg = stat.mean(self._results[0])
        print(
            "The average reward was: "
            + str(avg)
            + ". Agents terminated: "
            + str(self._results[1][0])
            + ". Agents with neg reward: "
            + str(self._results[1][3])
            + ". Agents truncated: "
            + str(self._results[1][1])
            + ". Agents over max_steps: "
            + str(self._results[1][2])
            + "."
        )


def evaluate_multiple_agents(env, agents, number, max_steps=100, GAMMA=0.99):
    result = ([], [0, 0, 0, 0])

    for agent in agents:
        env.reset()
        evaluator = ModularEvaluator(env, agent)
        current_result = evaluator.evaluate(number, max_steps, GAMMA)

        result[0].extend(current_result[0])
        result[1][0] += current_result[1][0]
        result[1][1] += current_result[1][1]
        result[1][2] += current_result[1][2]
        result[1][3] += current_result[1][3]

    return result



# a dictionary mapping exits [north, east, south, west] to the corresponding tile represented by a 2d list
TILES: dict[tuple[int, int, int, int], list[list[set[str]]]] = {
    (0, 0, 0, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 0, 0, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 1, 0, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 0, 1, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 0, 0, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 0, 1, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 1, 0, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 1, 0, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 1, 1, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [{"wall"}, {"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 0, 1, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 0, 0, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}, {"wall"}],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 1, 1, 0): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 1, 0, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}, {"wall"}],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 0, 1, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (0, 1, 1, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [{"wall"}, {"wall"}, {"wall"}, set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
    (1, 1, 1, 1): [
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit west"},
            {"exit west"},
            {"exit west"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [
            {"exit north"},
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            set(),
            {"exit south"},
        ],
        [{"wall"}, {"wall"}, set(), set(), set(), set(), set(), {"wall"}, {"wall"}],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            set(),
            set(),
            set(),
            {"wall"},
            {"wall"},
            {"wall"},
        ],
        [
            {"wall"},
            {"wall"},
            {"wall"},
            {"exit east"},
            {"exit east"},
            {"exit east"},
            {"wall"},
            {"wall"},
            {"wall"},
        ],
    ],
}

# a dictionary mapping the name of obstacle masks to their representation by a 2d list
OBSTACLE_MASKS: dict[str, list[list[set[str]]]] = {
    "blob": [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    "small_blob": [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), {"obstacle"}, set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), {"obstacle"}, set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    "bottom_half": [
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
    ],
    "top_half": [
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            set(),
        ],
    ],
    "left_half": [
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    "right_half": [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
    ],
    "chess_field": [
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
    ],
    "reverse_chess_field": [
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
        [
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
            set(),
            set(),
            set(),
            {"obstacle"},
            {"obstacle"},
            {"obstacle"},
        ],
    ],
}

# a dictionary mapping exits [north, east, south, west] to the corresponding traffic lanes represented by a 2d list
TRAFFIC_LANES: dict[tuple[int, int, int, int], list[list[set[str]]]] = {
    (0, 0, 0, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left"},
            set(),
            {"car_lane all right", "car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left"},
            set(),
            {"car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left"},
            set(),
            {"car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left"},
            set(),
            {"car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left"},
            set(),
            {"car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_west left", "car_spawner"},
            set(),
            {"car_lane west_to_middle right"},
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (0, 0, 1, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            {"car_lane middle_to_south down", "car_spawner"},
            {"car_lane middle_to_south down"},
            {"car_lane middle_to_south down"},
            {"car_lane middle_to_south down"},
            {"car_lane middle_to_south down"},
            {"car_lane middle_to_south down"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_middle up"},
            {"car_lane south_to_middle up"},
            {"car_lane south_to_middle up"},
            {"car_lane south_to_middle up"},
            {"car_lane south_to_middle up"},
            {"car_lane all up", "car_lane south_to_middle up"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (0, 1, 0, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right", "car_spawner"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane all left", "car_lane east_to_middle left"},
            set(),
            {"car_lane middle_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 0, 0, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"car_lane all down", "car_lane north_to_middle down"},
            {"car_lane north_to_middle down"},
            {"car_lane north_to_middle down"},
            {"car_lane north_to_middle down"},
            {"car_lane north_to_middle down"},
            {"car_lane north_to_middle down"},
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"car_lane middle_to_north up"},
            {"car_lane middle_to_north up"},
            {"car_lane middle_to_north up"},
            {"car_lane middle_to_north up"},
            {"car_lane middle_to_north up"},
            {"car_lane middle_to_north up", "car_spawner"},
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (0, 0, 1, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west left"},
            set(),
            {"car_lane all right", "car_lane west_to_south right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west left"},
            set(),
            {"car_lane west_to_south right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west left"},
            set(),
            {"car_lane west_to_south right"},
            {"car_lane west_to_south down"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west left"},
            set(),
            set(),
            {"car_lane west_to_south right"},
            {"car_lane west_to_south down"},
            {"car_lane west_to_south down"},
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west up"},
            {"car_lane south_to_west left"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"car_lane south_to_west up"},
            {"car_lane south_to_west up"},
            {"car_lane south_to_west up"},
            {"car_lane south_to_west up"},
            {"car_lane all up", "car_lane south_to_west up"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (0, 1, 0, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane all right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane all left", "car_lane east_to_west left"},
            set(),
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 0, 0, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane north_to_west left"},
            set(),
            {"car_lane all right", "car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane north_to_west left"},
            set(),
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane north_to_west left"},
            {"car_lane north_to_west down"},
            set(),
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane all down", "car_lane north_to_west down"},
            {"car_lane north_to_west down"},
            {"car_lane north_to_west down"},
            set(),
            set(),
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"car_lane west_to_north up"},
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane west_to_north up"},
            {"car_lane west_to_north up"},
            {"car_lane west_to_north up"},
            {"car_lane west_to_north up"},
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
            set(),
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (0, 1, 1, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            set(),
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            {"car_lane east_to_south down"},
            {"car_lane east_to_south down"},
            {"car_lane east_to_south down"},
            {"car_lane east_to_south down"},
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            {"car_lane east_to_south down"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            set(),
            set(),
            {"car_lane south_to_east up"},
            {"car_lane south_to_east up"},
            {"car_lane all up", "car_lane south_to_east up"},
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            set(),
            {"car_lane south_to_east up"},
            {"car_lane south_to_east right"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            set(),
            {"car_lane south_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane all left", "car_lane east_to_south left"},
            set(),
            {"car_lane south_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 0, 1, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"car_lane all down", "car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane all up", "car_lane south_to_north up"},
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (1, 1, 0, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {"car_lane all down", "car_lane north_to_east down"},
            {"car_lane north_to_east down"},
            {"car_lane north_to_east down"},
            {"car_lane north_to_east down"},
            {"car_lane north_to_east down"},
            set(),
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            set(),
            {"car_lane north_to_east right"},
            {"car_lane north_to_east down"},
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane east_to_north up"},
            {"car_lane east_to_north up"},
            {"car_lane east_to_north left"},
            set(),
            set(),
            {"car_lane north_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane east_to_north up"},
            {"car_lane east_to_north left"},
            set(),
            {"car_lane north_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_north left"},
            set(),
            {"car_lane north_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane all left", "car_lane east_to_north left"},
            set(),
            {"car_lane north_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (0, 1, 1, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane south_to_west left"},
            set(),
            {
                "car_lane all right",
                "car_lane west_to_east right",
                "car_lane west_to_south right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane south_to_west left"},
            set(),
            {"car_lane west_to_east right", "car_lane west_to_south right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane south_to_west left"},
            set(),
            {"car_lane west_to_east right", "car_lane west_to_south right"},
            {"car_lane west_to_south down"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane south_to_west left"},
            {"car_lane east_to_south left"},
            {"car_lane east_to_south down", "car_lane west_to_east right"},
            {"car_lane east_to_south down", "car_lane west_to_south right"},
            {"car_lane east_to_south down", "car_lane west_to_south down"},
            {"car_lane east_to_south down", "car_lane west_to_south down"},
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane east_to_south left",
                "car_lane east_to_west left",
                "car_lane south_to_west up",
            },
            {"car_lane east_to_south down", "car_lane south_to_west left"},
            {"car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left", "car_lane east_to_west left"},
            {"car_lane south_to_west up"},
            {"car_lane south_to_west up", "car_lane west_to_east right"},
            {"car_lane south_to_east up", "car_lane south_to_west up"},
            {"car_lane south_to_east up", "car_lane south_to_west up"},
            {
                "car_lane all up",
                "car_lane south_to_east up",
                "car_lane south_to_west up",
            },
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left", "car_lane east_to_west left"},
            set(),
            {"car_lane south_to_east up", "car_lane west_to_east right"},
            {"car_lane south_to_east right"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left", "car_lane east_to_west left"},
            set(),
            {"car_lane south_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane all left",
                "car_lane east_to_south left",
                "car_lane east_to_west left",
            },
            set(),
            {"car_lane south_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 0, 1, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane north_to_west left", "car_lane south_to_west left"},
            set(),
            {
                "car_lane all right",
                "car_lane west_to_north right",
                "car_lane west_to_south right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane north_to_west left", "car_lane south_to_west left"},
            set(),
            {"car_lane west_to_north right", "car_lane west_to_south right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane north_to_west left"},
            {"car_lane north_to_west down", "car_lane south_to_west left"},
            set(),
            {"car_lane west_to_north right", "car_lane west_to_south right"},
            {"car_lane west_to_south down"},
            set(),
            set(),
        ],
        [
            {
                "car_lane all down",
                "car_lane north_to_south down",
                "car_lane north_to_west down",
            },
            {"car_lane north_to_south down", "car_lane north_to_west down"},
            {"car_lane north_to_south down", "car_lane north_to_west down"},
            {"car_lane north_to_south down", "car_lane south_to_west left"},
            {"car_lane north_to_south down"},
            {"car_lane north_to_south down", "car_lane west_to_north right"},
            {"car_lane north_to_south down", "car_lane west_to_south right"},
            {"car_lane north_to_south down", "car_lane west_to_south down"},
            {"car_lane north_to_south down", "car_lane west_to_south down"},
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane south_to_west up"},
            {"car_lane south_to_west left", "car_lane west_to_north up"},
            {"car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane south_to_north up", "car_lane west_to_north up"},
            {"car_lane south_to_north up", "car_lane west_to_north up"},
            {"car_lane south_to_north up", "car_lane west_to_north up"},
            {"car_lane south_to_north up", "car_lane west_to_north up"},
            {
                "car_lane south_to_north up",
                "car_lane south_to_west up",
                "car_lane west_to_north right",
            },
            {"car_lane south_to_north up", "car_lane south_to_west up"},
            {"car_lane south_to_north up", "car_lane south_to_west up"},
            {"car_lane south_to_north up", "car_lane south_to_west up"},
            {
                "car_lane all up",
                "car_lane south_to_north up",
                "car_lane south_to_west up",
            },
        ],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
    ],
    (1, 1, 0, 1): [
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane north_to_west left"},
            set(),
            {
                "car_lane all right",
                "car_lane west_to_east right",
                "car_lane west_to_north right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left", "car_lane north_to_west left"},
            set(),
            {"car_lane west_to_east right", "car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane north_to_west left"},
            {"car_lane east_to_west left", "car_lane north_to_west down"},
            set(),
            {"car_lane west_to_east right", "car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            {
                "car_lane all down",
                "car_lane north_to_east down",
                "car_lane north_to_west down",
            },
            {"car_lane north_to_east down", "car_lane north_to_west down"},
            {"car_lane north_to_east down", "car_lane north_to_west down"},
            {"car_lane east_to_west left", "car_lane north_to_east down"},
            {"car_lane north_to_east down"},
            {"car_lane west_to_east right", "car_lane west_to_north right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_west left"},
            {"car_lane north_to_east right", "car_lane west_to_north up"},
            {
                "car_lane north_to_east down",
                "car_lane west_to_east right",
                "car_lane west_to_north right",
            },
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane east_to_north up", "car_lane west_to_north up"},
            {"car_lane east_to_north up", "car_lane west_to_north up"},
            {"car_lane east_to_north left", "car_lane west_to_north up"},
            {"car_lane east_to_west left", "car_lane west_to_north up"},
            {"car_lane west_to_north right"},
            {"car_lane north_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane east_to_north up"},
            {"car_lane east_to_north left", "car_lane east_to_west left"},
            set(),
            {"car_lane north_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_north left", "car_lane east_to_west left"},
            set(),
            {"car_lane north_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane all left",
                "car_lane east_to_north left",
                "car_lane east_to_west left",
            },
            set(),
            {"car_lane north_to_east right", "car_lane west_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 1, 1, 0): [
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [set(), set(), set(), set(), set(), set(), set(), set(), set()],
        [
            {
                "car_lane all down",
                "car_lane north_to_east down",
                "car_lane north_to_south down",
            },
            {"car_lane north_to_east down", "car_lane north_to_south down"},
            {"car_lane north_to_east down", "car_lane north_to_south down"},
            {"car_lane north_to_east down", "car_lane north_to_south down"},
            {
                "car_lane east_to_south left",
                "car_lane north_to_east down",
                "car_lane north_to_south down",
            },
            {"car_lane east_to_south down", "car_lane north_to_south down"},
            {"car_lane east_to_south down", "car_lane north_to_south down"},
            {"car_lane east_to_south down", "car_lane north_to_south down"},
            {"car_lane east_to_south down", "car_lane north_to_south down"},
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_south left"},
            {"car_lane east_to_south down", "car_lane north_to_east right"},
            {"car_lane north_to_east down"},
            set(),
            set(),
            set(),
        ],
        [
            {"car_lane east_to_north up", "car_lane south_to_north up"},
            {"car_lane east_to_north up", "car_lane south_to_north up"},
            {"car_lane east_to_north left", "car_lane south_to_north up"},
            {"car_lane east_to_south left", "car_lane south_to_north up"},
            {"car_lane south_to_north up"},
            {"car_lane north_to_east right", "car_lane south_to_north up"},
            {"car_lane south_to_east up", "car_lane south_to_north up"},
            {"car_lane south_to_east up", "car_lane south_to_north up"},
            {
                "car_lane all up",
                "car_lane south_to_east up",
                "car_lane south_to_north up",
            },
        ],
        [
            set(),
            set(),
            {"car_lane east_to_north up"},
            {"car_lane east_to_north left", "car_lane east_to_south left"},
            set(),
            {"car_lane north_to_east right", "car_lane south_to_east up"},
            {"car_lane south_to_east right"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {"car_lane east_to_north left", "car_lane east_to_south left"},
            set(),
            {"car_lane north_to_east right", "car_lane south_to_east right"},
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane all left",
                "car_lane east_to_north left",
                "car_lane east_to_south left",
            },
            set(),
            {"car_lane north_to_east right", "car_lane south_to_east right"},
            set(),
            set(),
            set(),
        ],
    ],
    (1, 1, 1, 1): [
        [
            set(),
            set(),
            set(),
            {
                "car_lane east_to_west left",
                "car_lane north_to_west left",
                "car_lane south_to_west left",
            },
            set(),
            {
                "car_lane all right",
                "car_lane west_to_east right",
                "car_lane west_to_north right",
                "car_lane west_to_south right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane east_to_west left",
                "car_lane north_to_west left",
                "car_lane south_to_west left",
            },
            set(),
            {
                "car_lane west_to_east right",
                "car_lane west_to_north right",
                "car_lane west_to_south right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            {"car_lane north_to_west left"},
            {
                "car_lane east_to_west left",
                "car_lane north_to_west down",
                "car_lane south_to_west left",
            },
            set(),
            {
                "car_lane west_to_east right",
                "car_lane west_to_north right",
                "car_lane west_to_south right",
            },
            {"car_lane west_to_south down"},
            set(),
            set(),
        ],
        [
            {
                "car_lane all down",
                "car_lane north_to_east down",
                "car_lane north_to_south down",
                "car_lane north_to_west down",
            },
            {
                "car_lane north_to_east down",
                "car_lane north_to_south down",
                "car_lane north_to_west down",
            },
            {
                "car_lane north_to_east down",
                "car_lane north_to_south down",
                "car_lane north_to_west down",
            },
            {
                "car_lane east_to_west left",
                "car_lane north_to_east down",
                "car_lane north_to_south down",
                "car_lane south_to_west left",
            },
            {
                "car_lane east_to_south left",
                "car_lane north_to_east down",
                "car_lane north_to_south down",
            },
            {
                "car_lane east_to_south down",
                "car_lane north_to_south down",
                "car_lane west_to_east right",
                "car_lane west_to_north right",
            },
            {
                "car_lane east_to_south down",
                "car_lane north_to_south down",
                "car_lane west_to_south right",
            },
            {
                "car_lane east_to_south down",
                "car_lane north_to_south down",
                "car_lane west_to_south down",
            },
            {
                "car_lane east_to_south down",
                "car_lane north_to_south down",
                "car_lane west_to_south down",
            },
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane east_to_south left",
                "car_lane east_to_west left",
                "car_lane south_to_west up",
            },
            {
                "car_lane east_to_south down",
                "car_lane north_to_east right",
                "car_lane south_to_west left",
                "car_lane west_to_north up",
            },
            {
                "car_lane north_to_east down",
                "car_lane west_to_east right",
                "car_lane west_to_north right",
            },
            set(),
            set(),
            set(),
        ],
        [
            {
                "car_lane east_to_north up",
                "car_lane south_to_north up",
                "car_lane west_to_north up",
            },
            {
                "car_lane east_to_north up",
                "car_lane south_to_north up",
                "car_lane west_to_north up",
            },
            {
                "car_lane east_to_north left",
                "car_lane south_to_north up",
                "car_lane west_to_north up",
            },
            {
                "car_lane east_to_south left",
                "car_lane east_to_west left",
                "car_lane south_to_north up",
                "car_lane west_to_north up",
            },
            {
                "car_lane south_to_north up",
                "car_lane south_to_west up",
                "car_lane west_to_north right",
            },
            {
                "car_lane north_to_east right",
                "car_lane south_to_north up",
                "car_lane south_to_west up",
                "car_lane west_to_east right",
            },
            {
                "car_lane south_to_east up",
                "car_lane south_to_north up",
                "car_lane south_to_west up",
            },
            {
                "car_lane south_to_east up",
                "car_lane south_to_north up",
                "car_lane south_to_west up",
            },
            {
                "car_lane all up",
                "car_lane south_to_east up",
                "car_lane south_to_north up",
                "car_lane south_to_west up",
            },
        ],
        [
            set(),
            set(),
            {"car_lane east_to_north up"},
            {
                "car_lane east_to_north left",
                "car_lane east_to_south left",
                "car_lane east_to_west left",
            },
            set(),
            {
                "car_lane north_to_east right",
                "car_lane south_to_east up",
                "car_lane west_to_east right",
            },
            {"car_lane south_to_east right"},
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane east_to_north left",
                "car_lane east_to_south left",
                "car_lane east_to_west left",
            },
            set(),
            {
                "car_lane north_to_east right",
                "car_lane south_to_east right",
                "car_lane west_to_east right",
            },
            set(),
            set(),
            set(),
        ],
        [
            set(),
            set(),
            set(),
            {
                "car_lane all left",
                "car_lane east_to_north left",
                "car_lane east_to_south left",
                "car_lane east_to_west left",
            },
            set(),
            {
                "car_lane north_to_east right",
                "car_lane south_to_east right",
                "car_lane west_to_east right",
            },
            set(),
            set(),
            set(),
        ],
    ],
}



from dataclasses import dataclass
from typing import Any

import graph

from pgtg.constants import OBSTACLE_MASK_NAMES, OBSTACLE_NAMES


@dataclass
class MapPlan:
    """Dataclass representing a map that has been generated but can't be used for running an episode yet."""

    width: int
    height: int
    tiles: list[list[dict[str, Any]]]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "MapPlan":
        """Creates a MapPlan object from a dictionary."""

        return cls(
            width=data["width"],
            height=data["height"],
            tiles=data["map"],
        )


def generate_map(
    width: int,
    height: int,
    percentage_of_connections: float,
    rng,
    *,
    obstacle_probability: float = 0,
    ice_probability_weight: float = 1,
    broken_road_probability_weight: float = 1,
    sand_probability_weight: float = 1,
) -> MapPlan:
    """Generates a map object based on the provided parameters.

    Args:
        width: The width of the generated map object in tiles.
        height: The height of the generated map object in tiles.
        percentage_of_connections: The percentage of connections that are generated compared to all possible connection that could exist. A value of 1 lead each tile being connected to all neighbors. The start and end of the map are always connected, even if a value of 0 is chosen.
        rng: (np rng) A rng that is used for all randomness.
        obstacle_probability: Default 0. The probability of a tile having a obstacle.
        ice_probability_weight: Default 1. Relative weight of the ice obstacle when choosing a random obstacle.
        broken_road_probability_weight: Default 1. Relative weight of the broken road obstacle when choosing a random obstacle.
        sand_probability_weight: Default 1. Relative weight of the sand obstacle when choosing a random obstacle.

    Returns:
        A object representing a map.
    """
    map_graph = generate_map_graph(width, height, percentage_of_connections, rng)
    tile_map_object = map_graph_to_tile_map_object(width, height, map_graph)
    add_connections_to_borders(tile_map_object, percentage_of_connections, rng)

    if obstacle_probability > 0:
        add_obstacles_to_map(
            tile_map_object,
            obstacle_probability,
            rng,
            ice_probability_weight=ice_probability_weight,
            broken_road_probability_weight=broken_road_probability_weight,
            sand_probability_weight=sand_probability_weight,
        )

    return tile_map_object


def generate_map_graph(
    width: int, height: int, percentage_of_connections: float, rng
) -> graph.Graph:
    """Generates a graph representing a map based on the provided parameters.

    Args:
        width: The width of the generated map graph in tiles.
        height: The height of the generated map graph in tiles.
        percentage_of_connections: The percentage of connections that are generated compared to all possible connection that could exist. A value of 1 lead each tile being connected to all neighbors. The start and end of the map are always connected, even if a value of 0 is chosen.
        rng: (np rng) A rng that is used for all randomness.

    Returns:
        A graph representing a map.
    """

    map_graph = graph.Graph()

    for i in range(width):
        for j in range(height):
            current_node_id = i + j * width

            if not (i == width - 1):
                map_graph.add_edge(current_node_id, current_node_id + 1, 1, True)
            if not (j == height - 1):
                map_graph.add_edge(current_node_id, current_node_id + width, 1, True)

    removable_edges = [(edge[0], edge[1]) for edge in map_graph.edges()]

    map_graph.add_edge("start", 0, 1, True)
    map_graph.add_edge("end", width * height - 1, 1, True)

    num_edges_to_keep = round(len(removable_edges) * percentage_of_connections)

    shortest_path = map_graph.breadth_first_search("start", "end")

    # remove edges until the desired number of edges is reached or no more edges can be removed (don't count the four edges that are always there: start to first node and reverse, end to last node and reverse)
    while len(map_graph.edges()) - 4 > num_edges_to_keep and len(removable_edges) > 0:

        chosen_edge = tuple(rng.choice(removable_edges))
        chosen_edge_reverse = tuple(reversed(chosen_edge))

        removable_edges.remove(chosen_edge)
        removable_edges.remove(chosen_edge_reverse)

        map_graph.del_edge(*chosen_edge)
        map_graph.del_edge(*chosen_edge_reverse)

        if all(x in shortest_path for x in chosen_edge):
            # it is not necessary to check if the two nodes of the chosen edge appear behind each other in the shortest path, if they didn't it wouldn't be a shortest path
            if map_graph.is_connected("start", "end"):
                shortest_path = map_graph.breadth_first_search("start", "end")
            else:
                map_graph.add_edge(*chosen_edge)
                map_graph.add_edge(*chosen_edge_reverse)

    return map_graph


def map_graph_to_tile_map_object(
    width: int, height: int, graph: graph.Graph
) -> MapPlan:
    """Turns a graph representing a map into a dict object representing the same map.

    Args:
        width: The width of the map in tiles.
        height: The height of the map in tiles.
        graph: The graph representing the map.

    Returns:
        The map object.
    """

    map_plan = MapPlan(width, height, [])

    for i in range(height - 1, -1, -1):
        row = []
        for j in range(width):
            tile = {}
            tile["exits"] = [0, 0, 0, 0]
            current_node_id = j + i * width
            neighboring_nodes = graph.nodes(from_node=current_node_id)

            if neighboring_nodes == None:
                continue

            if (current_node_id + width) in neighboring_nodes:  # tile to the north
                tile["exits"][0] = 1

            if (current_node_id + 1) in neighboring_nodes:  # tile to the east
                tile["exits"][1] = 1

            if (current_node_id - width) in neighboring_nodes:  # tile to the south
                tile["exits"][2] = 1

            if (current_node_id - 1) in neighboring_nodes:  # tile to the west
                tile["exits"][3] = 1

            row.append(tile)
        map_plan.tiles.append(row)

    map_plan.tiles[height - 1][0]["exits"][3] = 1  # add the start exit
    map_plan.tiles[0][width - 1]["exits"][1] = 1  # add the end exit

    return map_plan


def add_connections_to_borders(
    map: MapPlan, percentage_of_connections_to_edges: float, rng
) -> None:
    """Given a map object adds connections from the tiles next to the borders to the borders.

    Args:
        map_object: The map object that the connections will be added to.
        percentage_of_connections: The percentage of connections that are generated compared to all possible connection that could exist. A value of 1 lead each tile next to a border being connected to said border and a value of 0 leads to none being connected (except start and end).
        rng: (np rng) A rng that is used for all randomness.
    """

    width = map.width
    height = map.height

    # list of possible connections to edges as (tile_y, tile_x, direction) with direction: 0=north/top 1=east/right 2=south/bottom 3=west/left
    possible_connections_to_borders = (
        [(0, x, 0) for x in range(width)]  # connections to top edge
        + [(y, width - 1, 1) for y in range(height)]  # connections to right edge
        + [(height - 1, x, 2) for x in range(width)]  # connections to bottom edge
        + [(y, 0, 3) for y in range(height)]  # connection to left edge
    )

    possible_connections_to_borders.remove((height - 1, 0, 3))  # remove the start
    possible_connections_to_borders.remove((0, width - 1, 1))  # remove the goal

    num_connections_to_borders_to_add = round(
        len(possible_connections_to_borders) * percentage_of_connections_to_edges
    )

    for _ in range(num_connections_to_borders_to_add):
        connection_to_border_to_add = tuple(rng.choice(possible_connections_to_borders))
        possible_connections_to_borders.remove(connection_to_border_to_add)
        map.tiles[connection_to_border_to_add[0]][connection_to_border_to_add[1]][
            "exits"
        ][connection_to_border_to_add[2]] = 1


def add_obstacles_to_map(
    map: MapPlan,
    obstacle_probability: float,
    rng,
    *,
    ice_probability_weight: float = 1,
    broken_road_probability_weight: float = 1,
    sand_probability_weight: float = 1,
) -> None:
    """Given a map object adds obstacles to the tiles.

    Args:
        map_object: The map object that the obstacles will be added to.
        obstacle_probability: The probability of adding a obstacle to a tile.
        rng: (np rng) A rng that is used for all randomness.
        ice_probability_weight: Default 1. Relative weight of the ice obstacle when choosing a random obstacle.
        broken_road_probability_weight: Default 1. Relative weight of the broken road obstacle when choosing a random obstacle.
        sand_probability_weight: Default 1. Relative weight of the sand obstacle when choosing a random obstacle.
    """

    probability_weight_sum = (
        ice_probability_weight
        + broken_road_probability_weight
        + sand_probability_weight
    )

    ice_relative_probability_weight = ice_probability_weight / probability_weight_sum
    broken_road_relative_probability_weight = (
        broken_road_probability_weight / probability_weight_sum
    )
    sand_relative_probability_weight = sand_probability_weight / probability_weight_sum

    for row in range(map.height):
        for column in range(map.width):
            if (
                rng.random() < obstacle_probability
                and not str(map.tiles[row][column]["exits"]) == "[0, 0, 0, 0]"
            ):
                obstacle_type = rng.choice(
                    OBSTACLE_NAMES,
                    p=[
                        ice_relative_probability_weight,
                        broken_road_relative_probability_weight,
                        sand_relative_probability_weight,
                    ],
                )
                map.tiles[row][column]["obstacle_type"] = obstacle_type
                map.tiles[row][column]["obstacle_mask"] = rng.choice(
                    OBSTACLE_MASK_NAMES
                )



import json

from pgtg import parser
from pgtg.constants import DIRECTIONS_TO_INTS, TILE_HEIGHT, TILE_WIDTH
from pgtg.map_generator import MapPlan


class EpisodeMap:
    """Class representing a map for the environment that can be modified and used for running a episode."""

    def __init__(self, map_plan: MapPlan):
        # save the original map plan to be able to later save the map to a file
        self.map_plan = map_plan

        (
            self.width,
            self.height,
            self._map,
            self.num_subgoals,
            self.tile_coordinates_to_subgoal_directions,
        ) = parser.parse_map_object(self.map_plan)

        self.tile_width = int(self.width / TILE_WIDTH)
        self.tile_height = int(self.height / TILE_HEIGHT)

        self.starters = []
        self.goals = []
        self.traffic_spawnable_positions = []
        self.car_spawners = []

        for x in range(self.width):
            for y in range(self.height):
                if self.feature_at(x, y, "start"):
                    self.starters.append((x, y))
                if any(
                    ["car_lane" in feature for feature in self.get_features_at(x, y)]
                ):
                    self.traffic_spawnable_positions.append((x, y))
                if self.feature_at(x, y, "car_spawner"):
                    self.car_spawners.append((x, y))
                if self.feature_at(x, y, "final goal"):
                    self.goals.append((x, y))

    def inside_map(self, x: int, y: int) -> bool:
        """Returns true if the position specified by the x and y coordinates is inside the map and false otherwise."""

        return not (x < 0 or y < 0 or x >= self.width or y >= self.height)

    def get_features_at(self, x: int, y: int) -> set[str]:
        """Returns a list of all features at the position specified by the x and y coordinates."""

        if x < 0 or x > self.width - 1 or y < 0 or y > self.height - 1:
            raise ValueError("coordinates are outside the map")
        return self._map[x][y]

    def set_features_at(self, x: int, y: int, features: set[str]) -> None:
        """Replaces the features at the the position specified by the x and y coordinates with the provided ones."""

        if x < 0 or x > self.width - 1 or y < 0 or y > self.height - 1:
            raise ValueError("coordinates are outside the map")
        self._map[x][y] = features

    def feature_at(self, x: int, y: int, features: str | set[str] | list[str]) -> bool:
        """Returns true if one of the specified features is present at the the position specified by the x and y coordinates and false otherwise."""

        if isinstance(features, str):
            return features in self.get_features_at(x, y)
        else:
            return not self.get_features_at(x, y).isdisjoint(features)

    def add_feature_at(self, x: int, y: int, feature: str) -> None:
        """Adds the specified feature at the the position specified by the x and y coordinates."""
        self.get_features_at(x, y).add(feature)

    def remove_feature_at(self, x: int, y: int, feature: str) -> None:
        """Removes the specified feature at the the position specified by the x and y coordinates. Doesn't raise an error if feature is not present."""

        self.get_features_at(x, y).discard(feature)

    def get_map_cutout(
        self,
        top_left_x: int,
        top_left_y: int,
        bottom_right_x: int,
        bottom_right_y: int,
        fill_squares_outside_map_with: set[str] | None = None,
    ) -> list[list[set[str]]]:
        """Returns a rectangular cutout of the map. If parts of the cutout are outside the map, those squares are optionally filled with the specified features.

        Args:
            top_left_x: The x-value of the top left corner of the cutout.
            top_left_y: The y-value of the top left corner of the cutout.
            bottom_right_x: The x-value of the bottom right corner of the cutout.
            bottom_right_y: The y-value of the bottom right corner of the cutout.
            fill_squares_outside_map_with: Default None. Squares outside the map are filled with these features. If None, the squares are left empty.

        Returns:
            A cutout from the map.
        """

        cutout = [
            [set() for _ in range(bottom_right_y - top_left_y + 1)]
            for _ in range(bottom_right_x - top_left_x + 1)
        ]

        for x in range(top_left_x, bottom_right_x + 1):
            for y in range(top_left_y, bottom_right_y + 1):
                if self.inside_map(x, y):
                    cutout[x - top_left_x][y - top_left_y] = self.get_features_at(x, y)
                elif fill_squares_outside_map_with is not None:
                    cutout[x - top_left_x][
                        y - top_left_y
                    ] = fill_squares_outside_map_with

        return cutout

    def get_next_subgoal_direction(self, x: int, y: int) -> int:
        """Returns the direction of the next subgoal from the specified position or -1 if no subgoal is on that tile.

        Args:
            x: x-value of the position.
            y: y-value of the position.

        Returns:
            The direction of the next subgoal. (0: north, 1: east, 2: south, 3: west, -1: no subgoal found)
        """

        tile_x = int(x / TILE_WIDTH)
        tile_y = int(y / TILE_HEIGHT)

        direction = self.tile_coordinates_to_subgoal_directions.get(
            (tile_x, tile_y), None
        )

        if direction is None:
            return -1

        return DIRECTIONS_TO_INTS[direction]

    def set_subgoals_to_used(self, x: int, y: int) -> None:
        """Marks the subgoal at the specified coordinates as used and recursively marks all directly adjacent subgoals as used as well.

        Args:
            x: x-value of the position
            y: y-value of the position
        """

        assert self.feature_at(x, y, "subgoal"), (
            "Subgoal expected but found "
            + str(self.get_features_at(x, y))
            + " instead."
        )

        self.remove_feature_at(x, y, "subgoal")
        self.add_feature_at(x, y, "used subgoal")

        # recursively replace directly adjacent subgoal parts
        if self.feature_at(x, y + 1, "subgoal"):
            self.set_subgoals_to_used(x, y + 1)

        if self.feature_at(x, y - 1, "subgoal"):
            self.set_subgoals_to_used(x, y - 1)

        if self.feature_at(x + 1, y, "subgoal"):
            self.set_subgoals_to_used(x + 1, y)

        if self.feature_at(x - 1, y, "subgoal"):
            self.set_subgoals_to_used(x - 1, y)

    def save_map(self, path: str) -> None:
        """Saves the map as a JSON file.

        Args:
            path (String): path to the file the map should be saved in.
        """

        if not path.endswith(".json"):
            path += ".json"

        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.map_plan, f, ensure_ascii=False, indent=4)



# the width and height of a tile in squares
TILE_WIDTH: int = 9
TILE_HEIGHT: int = 9

# dictionary mapping action numbers to (x,y) acceleration
ACTIONS_TO_ACCELERATION: dict[int, tuple[int, int]] = {
    0: (-1, -1),  # left-up
    1: (-1, 0),  # left
    2: (-1, 1),  # left-down
    3: (0, -1),  # up
    4: (0, 0),  # nothing
    5: (0, 1),  # down
    6: (1, -1),  # right-up
    7: (1, 0),  # right
    8: (1, 1),  # right-down
}

OBSTACLE_NAMES: list[str] = ["ice", "broken road", "sand"]
OBSTACLE_MASK_NAMES: list[str] = [
    "blob",
    "small_blob",
    "chess_field",
    "reverse_chess_field",
    "top_half",
    "bottom_half",
    "left_half",
    "right_half",
]

# a dictionary mapping the names of cardinal directions to integers
DIRECTIONS_TO_INTS: dict[str, int] = {
    "north": 0,
    "east": 1,
    "south": 2,
    "west": 3,
}



import copy
from collections import OrderedDict
from typing import Any, SupportsFloat

import gymnasium as gym
import numpy as np
import numpy.typing as npt
import pygame
from gymnasium import spaces
from PIL.Image import Image

from pgtg import graphic
from pgtg.constants import ACTIONS_TO_ACCELERATION, TILE_HEIGHT, TILE_WIDTH
from pgtg.map import EpisodeMap
from pgtg.map_generator import generate_map
from pgtg.parser import json_file_to_map_plan


def _round(x):
    return int(np.floor(x + 0.5))


class PGTGEnv(gym.Env):
    """Class representing the modular racetrack environment."""

    metadata = {"render_modes": ["human", "rgb_array", "pil_image"], "render_fps": 4}

    def __init__(
        self,
        map_path: str | None = None,
        *,
        random_map_width: int = 4,
        random_map_height: int = 4,
        random_map_percentage_of_connections: float = 0.5,
        random_map_obstacle_probability: float = 0.0,
        random_map_ice_probability_weight: float = 1,
        random_map_broken_road_probability_weight: float = 1,
        random_map_sand_probability_weight: float = 1,
        render_mode: str | None = None,
        use_sliding_observation_window: bool = False,
        sliding_observation_window_size: int = 4,
        use_next_subgoal_direction: bool = False,
        sum_subgoals_reward: int = 100,
        final_goal_bonus: int = 0,
        crash_penalty: int = 100,
        standing_still_penalty: int = 0,
        already_visited_position_penalty: int = 0,
        ice_probability: float = 0.1,
        street_damage_probability: float = 0.1,
        sand_probability: float = 0.2,
        traffic_density: float = 0.0,
        ignore_traffic_collisions: bool = False,
    ):
        """initializes an object of the class with the given arguments.

        Args:
            map_name: Default None. Path to map file, either absolute or relative to the current working directory. If it is None a random map is generated instead.
            random_map_width: Default 4. Width of the random map.
            random_map_height: Default 4. Height of the random map.
            random_map_percentage_of_connections: Default 0.5. Percentage of connections between tiles that will be kept while generating the random map.
            random_map_obstacle_probability: Default 0.0. Probability that a tile receives an obstacle while generating the random map.
            random_map_ice_probability_weight: Default 1. Relative weight of the ice obstacle when generating the random map.
            random_map_broken_road_probability_weight: Default 1. Relative weight of the broken_road obstacle when generating the random map.
            random_map_sand_probability_weight: Default 1. Relative weight of the sand obstacle when generating the random map.
            render_mode: Default None. What kind of output render() generates. "human" results in no output by render() but automatic displaying during step().
            use_sliding_observation_window: Default False. Whether or not to use a observation window that moves with the agent. If False instead the current tile is observed.
            sliding_observation_window_size: Default 4. The distance to the border of the moving observation window. A size of 1 results in a 3x3 window, a size of 4 in a 9x9 window and a size of 0 in a 1x1 window
            (only the currently occupied square is visible). No effect if use_sliding_observation_window is False.
            use_next_subgoal_direction: Default False. Whether or not to use the direction to the next subgoal as an additional feature in the observation.
            sum_subgoals_reward: Default 100. The sum of all subgoal rewards. Gets split evenly among all subgoals.
            final_goal_bonus: Default 0. Bonus for the final goal that gets added to the subgoal reward.
            crash_penalty: Default 100. Penalty for driving into a wall or out of the map. The value will get negated so a positive value should be used.
            standing_still_penalty: Default 0. Penalty for not moving. Gets applied every step. The value will get negated so a positive value should be used.
            already_visited_position_penalty: Default 0. Penalty for moving to a position that was visited before. Gets applied every step for the final position of that step. Standing still does not result in penalty. The value will get negated so a positive value should be used.
            ice_probability: Default 0.1. Probability of the ice effect happening when the agent drives over it.
            street_damage_probability: Default  0.1. Probability of the broken road effect happening when the agent drives over it.
            sand_probability: Default 0.2. Probability of the sand effect happening when the agent drives over it.
            traffic_density: Default 0.0. The density of the traffic as fraction of the number of traffic cars and the number of positions where traffic could be. For a value of 0 no traffic is generated.
            ignore_traffic_collisions: Default False. If true collisions with traffic are ignored. For testing purposes.
        """

        # There are 8 different directions to accelerate into and the option to stand still.
        self.action_space = spaces.Discrete(9)

        observation_window_size = (
            (TILE_WIDTH, TILE_HEIGHT)
            if not use_sliding_observation_window
            else (
                1 + sliding_observation_window_size * 2,
                1 + sliding_observation_window_size * 2,
            )
        )

        # The agent sees the position in the current tile, its velocity and the map tile.
        observation_space_dict = {
            "position": spaces.MultiDiscrete([TILE_WIDTH, TILE_HEIGHT], dtype=np.int32),
            "velocity": spaces.Box(low=-99, high=99, shape=(2,), dtype=np.int32),
            "map": spaces.Dict(
                {
                    "walls": spaces.MultiBinary(observation_window_size),
                    "goals": spaces.MultiBinary(observation_window_size),
                    "ice": spaces.MultiBinary(observation_window_size),
                    "broken road": spaces.MultiBinary(observation_window_size),
                    "sand": spaces.MultiBinary(observation_window_size),
                    "traffic": spaces.MultiBinary(observation_window_size),
                }
            ),
        }

        if use_next_subgoal_direction:
            observation_space_dict["next_subgoal_direction"] = spaces.Discrete(
                5, start=-1
            )

        self.observation_space = spaces.Dict(observation_space_dict)

        self.render_mode = render_mode

        self.use_sliding_observation_window = use_sliding_observation_window
        self.sliding_observation_window_size = sliding_observation_window_size
        self.use_next_subgoal_direction = use_next_subgoal_direction

        self.reward_range = (-np.inf, np.inf)

        self.map_path = map_path
        self.map_plan = None

        self.random_map_width = random_map_width
        self.random_map_height = random_map_height
        self.random_map_percentage_of_connections = random_map_percentage_of_connections
        self.random_map_obstacle_probability = random_map_obstacle_probability
        self.random_map_ice_probability_weight = random_map_ice_probability_weight
        self.random_map_broken_road_probability_weight = (
            random_map_broken_road_probability_weight
        )
        self.random_map_sand_probability_weight = random_map_sand_probability_weight

        self.sum_subgoals_reward = sum_subgoals_reward  # the sum of all subgoal rewards, the individual subgoal reward is all_subgoals_reward / # of subgoals
        self.final_goal_bonus = final_goal_bonus
        self.crash_penalty = crash_penalty
        self.standing_still_penalty = standing_still_penalty
        self.already_visited_position_penalty = already_visited_position_penalty

        self.ice_probability = ice_probability
        self.street_damage_probability = street_damage_probability
        self.sand_probability = sand_probability
        self.traffic_density = traffic_density

        self.ignore_traffic_collisions = ignore_traffic_collisions

        self.window_size = 720
        self.window = None
        self.clock = None

    def reset(
        self, *, seed: int | None = None, options: dict[str, Any] | None = None
    ) -> tuple[OrderedDict, dict[str, Any]]:
        """Resets the environment. Must be used before starting a episode.

        Returns:
            observation: One element from within the observation space.
            info: Additional information about the state of the environment. Analogous to the info returned by step().
        """
        super().reset(seed=seed)

        # define one random number generator for each random element
        (
            self.map_rng,
            self.car_rng,
            self.ice_rng,
            self.broken_road_rng,
            self.sand_rng,
        ) = self.np_random.spawn(5)

        if (
            not self.map_path == None
        ):  # if a map_path is provided load the map from that path
            if self.map_plan == None:  # only load the file on the first reset
                self.map_plan = json_file_to_map_plan(self.map_path)
            self.map = EpisodeMap(self.map_plan)
        else:
            random_generated_map_object = generate_map(
                self.random_map_width,
                self.random_map_height,
                self.random_map_percentage_of_connections,
                self.map_rng,
                obstacle_probability=self.random_map_obstacle_probability,
                ice_probability_weight=self.random_map_ice_probability_weight,
                broken_road_probability_weight=self.random_map_broken_road_probability_weight,
                sand_probability_weight=self.random_map_sand_probability_weight,
            )
            self.map = EpisodeMap(random_generated_map_object)

        self.individual_subgoal_reward = (
            self.sum_subgoals_reward / self.map.num_subgoals
        )

        self.position = np.array(self.map_rng.choice(self.map.starters))

        self.velocity = np.array([0, 0])

        self.terminated = False
        self.truncated = False
        self.flat_tire = False

        self.positions_path = [list(self.position)]
        self.tile_path = [list(self.position)]
        self.noise_path = []

        self.cars = []
        self._next_car_id = 0

        if self.traffic_density > 0:
            self._create_initial_traffic()

        return (self.get_observation(), self.get_info())

    def _decompose_velocity(
        self, velocity: npt.NDArray | None = None
    ) -> list[npt.NDArray | None]:
        """Decomposes the velocity to all intermediate steps of length 1.

        Args:
            velocity: The velocity to decompose. If None, current velocity is used.

        Returns:
            The list of the individual steps.
        """

        if velocity is None:
            velocity = self.velocity

        dx = velocity[0]
        dy = velocity[1]

        # first compute how the complete velocity change accumulates over time steps
        # trivial case:
        if dx == 0 and dy == 0:
            return []

        res = []
        if dx == 0:
            # each possible y value
            m = np.sign(dy)  # evaluates to 1 or -1, dependent on dy>0 or dy<0
            for i in range(1, np.abs(dy) + 1):
                res.append((0, i * m))
        elif dy == 0:
            m = np.sign(dx)  # evaluates to 1 or -1, dependent of dx>0 or dx<0
            for i in range(1, np.abs(dx) + 1):
                res.append((i * m, 0))
        elif np.abs(dx) >= np.abs(dy):
            m_y = dy / np.abs(dx)
            m_x = np.sign(dx)
            for i in range(1, np.abs(dx) + 1):
                act_x = int(i * m_x)
                act_y = int(_round(i * m_y))
                res.append((act_x, act_y))
        elif np.abs(dx) < np.abs(dy):
            m_x = dx / np.abs(dy)
            m_y = np.sign(dy)
            for i in range(1, np.abs(dy) + 1):
                act_y = int(i * m_y)
                act_x = int(_round(i * m_x))
                res.append((act_x, act_y))

        pre = np.array([0, 0])

        # now deconstruct these changes into single steps by subtracting the velocity already applied
        for i, vel in enumerate(res):
            tmp = np.array(vel)
            vel = tmp - pre
            res[i] = vel
            pre = tmp

        return res

    def generate_frame(
        self,
        hide_positions: bool = False,
        show_observation_window: bool = True,
    ) -> Image:
        """Generates a image showing the current state of the environment.

        Args:
            hide_positions: Default False. Whether or not to hide the positions of the agent.
            show_observation_window: Default True. Whether or not to show the observation window.

        Returns:
            The generated image.
        """

        pic = graphic.create_map(
            self,
            show_path=(not hide_positions),
            show_observation_window=show_observation_window,
        )

        return pic

    def render(self) -> Image | npt.NDArray | None:
        """Potentially returns a rendered representation of the game according to the render mode.

        Returns:
            The rendered representation of the game. If the render mode is "human" or None, None is returned.
        """

        match self.render_mode:
            case None:
                return None
            case "human":
                return None
            case "rgb_array":
                return np.transpose(
                    np.asarray(self.generate_frame().convert("RGB")), axes=(1, 0, 2)
                )
            case "pil_image":
                return self.generate_frame()
            case _:
                raise Exception("the selected render_mode is not supported")

    def _render_frame_for_human(self) -> None:
        """Renders the current state of the environment in a window."""

        if self.window is None:
            pygame.init()
            pygame.display.init()
            self.window = pygame.display.set_mode(
                (
                    self.window_size * (self.map.tile_width / self.map.tile_height),
                    self.window_size,
                )
            )
            pygame.display.set_caption("PGTG")

        if self.clock is None:
            self.clock = pygame.time.Clock()

        pil_image = self.generate_frame()

        pygame_image = pygame.image.fromstring(
            pil_image.tobytes(), pil_image.size, pil_image.mode  # type: ignore
        ).convert()

        pygame_image = pygame.transform.scale(
            pygame_image,
            (
                self.window_size * (pil_image.size[0] / pil_image.size[1]),
                self.window_size,
            ),
        )

        self.window.blit(pygame_image, pygame_image.get_rect())
        pygame.event.pump()
        pygame.display.update()

        # add a delay to keep the framerate stable
        self.clock.tick(self.metadata["render_fps"])

    def _create_initial_traffic(self) -> None:
        """Creates a number of cars defined by the traffic density."""

        initial_car_positions = self.car_rng.choice(
            self.map.traffic_spawnable_positions,
            size=int(len(self.map.traffic_spawnable_positions) * self.traffic_density),
            replace=False,
        )

        initial_car_positions = [
            tuple(initial_car_position)
            for initial_car_position in initial_car_positions
        ]

        for initial_car_position in initial_car_positions:
            routes = [
                feature.split()[1]
                for feature in self.map.get_features_at(*initial_car_position)
                if "car_lane" in feature and "all" not in feature
            ]

            # Because the features are a set, their order once converted to a list can vary, even with the same seed. Sorting the list makes the environment deterministic again.
            routes.sort()

            assert (
                len(routes) > 0
            ), "a car was spawned on a field where no car lane was found"

            self.cars.append(
                {
                    "id": self._next_car_id,
                    "position": initial_car_position,
                    "route": self.car_rng.choice(routes),
                }
            )

            self._next_car_id += 1

    def _move_car(self, car: dict[str, Any]) -> dict[str, Any] | None:
        """Moves a car to a new position if possible.

        Args:
            car: The car to move.

        Returns:
            The moved car. If there are no possible movements, None is returned.
        """

        x, y = car["position"]
        possible_directions = [(x, y - 1), (x, y + 1), (x - 1, y), (x + 1, y)]
        possible_types = [
            "up",
            "down",
            "left",
            "right",
        ]

        for possible_position, type in zip(possible_directions, possible_types):
            if not self.map.inside_map(*possible_position):
                continue
                # ignore possible positions outside the map

            square_lanes = [
                feature
                for feature in self.map.get_features_at(*possible_position)
                if "car_lane" in feature
            ]

            lanes_for_all = [lane for lane in square_lanes if "all" in lane]
            if len(lanes_for_all) > 0 and type in lanes_for_all[0]:
                possible_routes = [
                    lane.split()[1] for lane in square_lanes if lane.split()[1] != "all"
                ]

                # Because the features are a set, their order once converted to a list can vary, even with the same seed. Sorting the list makes the environment deterministic again.
                possible_routes.sort()

                car["position"] = possible_position
                car["route"] = self.car_rng.choice(possible_routes)
                return car

            else:
                for lane in square_lanes:
                    if car["route"] != None and car["route"] in lane and type in lane:

                        car["position"] = possible_position
                        return car

    def _spawn_new_car(self) -> dict[str, Any]:
        """Creates a new car and returns it. It still has to be stored.

        Returns:
            The newly created car.
        """

        position = tuple(self.car_rng.choice(self.map.car_spawners))
        routes = [
            feature.split()[1]
            for feature in self.map.get_features_at(*position)
            if "car_lane" in feature and "all" not in feature
        ]

        # Because the features are a set, their order once converted to a list can vary, even with the same seed. Sorting the list makes the environment deterministic again.
        routes.sort()

        new_car = {
            "id": self._next_car_id,
            "position": position,
            "route": self.car_rng.choice(routes),
        }
        self._next_car_id += 1

        return new_car

    def step(
        self, action: int
    ) -> tuple[OrderedDict, SupportsFloat, bool, bool, dict[str, Any]]:
        """Center piece of this class. Performs the given action and returns the results.

        Args:
            action: The action to perform.

        Returns:
            observation: One element from within the observation space.
            reward: The reward for the action.
            terminated flag: Whether or not the episode has ended.
            truncated flag: Whether or not the episode has been truncated.
            info: Additional information about the state of the environment. Analogous to the info returned by reset().
        """

        # check whether the game is already done
        if self.terminated or self.truncated:
            raise RuntimeError("Already done, step has no further effect")

        # translate the action id to the acceleration
        acceleration = np.array(ACTIONS_TO_ACCELERATION[action])

        # cars move first
        new_cars = []
        for car in self.cars:
            moved_car = self._move_car(car)
            if moved_car != None:
                new_cars.append(moved_car)
            else:
                # the old car is effectively deleted because it is not added to new_cars
                # instead a new car is spawned
                new_cars.append(self._spawn_new_car())
        self.cars = new_cars

        # set start variables
        reward = 0
        current_position: npt.NDArray = copy.copy(self.position)

        # handle the velocity
        self.velocity = self.velocity + acceleration

        decomposed_velocity: list[npt.NDArray | None] = self._decompose_velocity()
        # a "stand still" check to also check the final tile of the step
        decomposed_velocity.append(None)

        # process the single steps
        while decomposed_velocity:
            velocity_part = decomposed_velocity.pop(0)

            current_position_x, current_position_y = current_position

            # case outside map, wall, or traffic
            if (
                not self.map.inside_map(current_position_x, current_position_y)
                or self.map.feature_at(current_position_x, current_position_y, "wall")
                or (
                    not self.ignore_traffic_collisions
                    and tuple(current_position)
                    in [car["position"] for car in self.cars]
                )
            ):
                reward -= self.crash_penalty
                self.terminated = True
                break

            # case goal
            if self.map.feature_at(
                current_position_x, current_position_y, "final goal"
            ):
                reward += self.individual_subgoal_reward + self.final_goal_bonus
                self.terminated = True
                break

            # case subgoal
            if self.map.feature_at(current_position_x, current_position_y, "subgoal"):
                reward += self.individual_subgoal_reward
                self.map.set_subgoals_to_used(current_position_x, current_position_y)

            # if the last step -> only checking for goal and wall, skip the rest
            if velocity_part is None:
                continue

            # case ice
            if (
                self.map.feature_at(current_position_x, current_position_y, "ice")
                and self.ice_rng.random() < self.ice_probability
            ):
                # pick a random action
                ice_action = self.ice_rng.choice(list(range(9)))
                ice_velocity = np.array(ACTIONS_TO_ACCELERATION[ice_action])
                velocity_part = ice_velocity
                self.noise_path.append(list(current_position))

            # case road_break
            if (
                self.map.feature_at(
                    current_position_x, current_position_y, "broken road"
                )
                and self.broken_road_rng.random() < self.street_damage_probability
            ):
                self.flat_tire = True
                self.noise_path.append(list(current_position))

            # case sand
            if (
                self.map.feature_at(current_position_x, current_position_y, "sand")
                and self.sand_rng.random() < self.sand_probability
            ):
                self.noise_path.append(list(current_position))
                current_position += velocity_part
                self.tile_path.append(list(current_position))
                self.velocity = np.array([0, 0])
                break

            current_position += velocity_part
            self.tile_path.append(list(current_position))

        # if there is a flat tire, the velocity is set to zero after each step.
        if self.flat_tire:
            self.velocity = np.array([0, 0])

        # apply penalty for moving to a already visited position
        if (
            self.already_visited_position_penalty != 0
            and not np.array_equal(acceleration, np.array([0, 0]))
            and any(
                [
                    np.array_equal(current_position, position_in_path)
                    for position_in_path in self.positions_path
                ]
            )
        ):
            reward -= self.already_visited_position_penalty

        # keep track of old and new position
        old_position = self.position
        self.position = current_position
        self.positions_path.append(list(self.position))

        if (
            self.standing_still_penalty != 0
            and np.array_equal(acceleration, np.array([0, 0]))
            and np.array_equal(old_position, current_position)
        ):
            reward -= self.standing_still_penalty

        if self.render_mode == "human":
            self._render_frame_for_human()

        # actually, the state is defined through pos and velocity and the distances are only features
        # for reasons of simpler implementation, the features here are returned together with the state
        return (
            self.get_observation(),
            reward,
            self.terminated,
            self.truncated,
            self.get_info(),
        )

    def light_step(
        self, action: int
    ) -> tuple[OrderedDict, SupportsFloat, bool, bool, dict[str, Any]]:
        """Copies the environment and executes a single step on it. The original environment remains unchanged.

        Args:
            action: The action to perform.

        Returns:
            observation: One element from within the observation space.
            reward: The reward for the action.
            terminated flag: Whether or not the episode has ended.
            truncated flag: Whether or not the episode has been truncated.
            info: Additional information about the state of the environment. Analogous to the info returned by reset().
        """

        env_copy = copy.deepcopy(self)
        return env_copy.step(action)

    def set_to_state(self, state: dict[str, Any]) -> tuple[OrderedDict, dict[str, Any]]:
        """Sets the environment to a given state.

        This function exists for easily making multiple recordings of a agents behavior starting from the same state.
        Setting two environments to the same state and choosing the same actions will NOT result in the same state afterwards, because the random number generators are not synchronized.

        Args:
            state: The state to set the environment to.

        Returns:
            observation: One element from within the observation space.
            info: Additional information about the state of the environment.
        """

        self.position[0] = state["x"]  # self.position[0] is the x coordinate
        self.position[1] = state["y"]  # self.position[1] is the y coordinate
        self.velocity[0] = state["x_velocity"]  # self.velocity[0] is the x velocity
        self.velocity[1] = state["y_velocity"]  # self.velocity[1] is the y velocity
        self.flat_tire = state["flat_tire"]

        self.cars = []
        if state["cars"] != None and len(state["cars"]) > 0:
            for car in state["cars"]:
                self.cars.append(
                    {
                        "id": car["id"],
                        "position": (car["x"], car["y"]),
                        "route": car["route"],
                    }
                )
            self._next_car_id = self.cars[-1]["id"] + 1

        return (self.get_observation(), self.get_info())

    def get_observation(self) -> OrderedDict[str, Any]:
        """Returns the current observation visible to the agent.

        Returns:
            A element from within the observation space.
        """

        # after the last step the agent could be outside the map
        position_inside_map_x = min(max(0, self.position[0]), self.map.width - 1)
        position_inside_map_y = min(max(0, self.position[1]), self.map.height - 1)

        tile_x = int(position_inside_map_x / TILE_WIDTH)
        tile_y = int(position_inside_map_y / TILE_HEIGHT)

        if not self.use_sliding_observation_window:
            cutout_top_left_x = tile_x * TILE_WIDTH
            cutout_top_left_y = tile_y * TILE_HEIGHT
            cutout_bottom_right_x = tile_x * TILE_WIDTH + TILE_WIDTH - 1
            cutout_bottom_right_y = tile_y * TILE_HEIGHT + TILE_HEIGHT - 1
        else:
            cutout_top_left_x = self.position[0] - self.sliding_observation_window_size
            cutout_top_left_y = self.position[1] - self.sliding_observation_window_size
            cutout_bottom_right_x = (
                self.position[0] + self.sliding_observation_window_size
            )
            cutout_bottom_right_y = (
                self.position[1] + self.sliding_observation_window_size
            )

        map_cutout = self.map.get_map_cutout(
            cutout_top_left_x,
            cutout_top_left_y,
            cutout_bottom_right_x,
            cutout_bottom_right_y,
            None if not self.use_sliding_observation_window else {"wall"},
        )

        map = OrderedDict(
            {
                "walls": np.array(self.encode_map_with_hot_one(map_cutout, "wall")),
                "goals": np.array(
                    self.encode_map_with_hot_one(map_cutout, {"subgoal", "final goal"})
                ),
                "ice": np.array(self.encode_map_with_hot_one(map_cutout, "ice")),
                "broken road": np.array(
                    self.encode_map_with_hot_one(map_cutout, "broken road")
                ),
                "sand": np.array(self.encode_map_with_hot_one(map_cutout, "sand")),
                "traffic": [],
            }
        )

        # creates a 2d array of the same size as map_cutout filled with 0
        traffic = np.array([[0] * len(map_cutout[0]) for _ in range(len(map_cutout))])

        for car in self.cars:
            if (
                cutout_top_left_x
                <= car["position"][0]  # car[0] is the car's x position
                <= cutout_bottom_right_x
                and cutout_top_left_y
                <= car["position"][1]  # car[1] is the car's y position
                <= cutout_bottom_right_y
            ):
                traffic[car["position"][0] - cutout_top_left_x][
                    car["position"][1] - cutout_top_left_y
                ] = 1

        map["traffic"] = traffic

        observation: OrderedDict[str, Any] = OrderedDict(
            {
                "position": np.array(
                    [
                        (
                            (position_inside_map_x - cutout_top_left_x)
                            if not self.use_sliding_observation_window
                            else 0
                        ),
                        (
                            (position_inside_map_y - cutout_top_left_y)
                            if not self.use_sliding_observation_window
                            else 0
                        ),
                    ]
                ),
                "velocity": self.velocity,
                "map": map,
            }
        )

        if self.use_next_subgoal_direction:
            next_subgoal_direction = self.map.get_next_subgoal_direction(
                position_inside_map_x, position_inside_map_y
            )
            observation["next_subgoal_direction"] = next_subgoal_direction

        return observation

    def encode_map_with_hot_one(
        self, map_cutout: list[list[set[str]]], features_to_match: str | set[str]
    ) -> list[list[int]]:
        """Transforms a map or map cutout into a hot-one encoding.
        If a square contains one or more of the features to match, the hot-one encoding will have a 1 at that position and otherwise it will have a 0.

        Args:
            map_cutout: The map or map cutout to transform.
            features_to_match: The feature(s) that will result in a 1 in the hot one encoding.

        Returns:
            The hot-one encoding for the specified feature(s).
        """

        assert isinstance(features_to_match, str) or isinstance(
            features_to_match, set
        ), "features_to_match must be a string or a set of strings"

        if isinstance(features_to_match, str):
            features_to_match = {features_to_match}

        # creates a 2d array of the same size as map_cutout filled with 0
        res = [[0] * len(map_cutout[0]) for _ in range(len(map_cutout))]

        for x in range(len(map_cutout)):
            for y in range(len(map_cutout[0])):
                if not map_cutout[x][y].isdisjoint(features_to_match):
                    res[x][y] = 1

        return res

    def get_info(self) -> dict[str, Any]:
        """Returns additional information about the state of the environment."""

        state = {
            "x": self.position[0],
            "y": self.position[1],
            "x_velocity": self.velocity[0],
            "y_velocity": self.velocity[1],
            "flat_tire": self.flat_tire,
            "cars": [],
        }
        for car in self.cars:
            state["cars"].append(
                {
                    "id": car["id"],
                    "x": car["position"][0],
                    "y": car["position"][1],
                    "route": car["route"],
                }
            )

        return state

    def applicable_actions(self) -> list[int]:
        """Returns list of applicable actions. For this environment it is always the same unless the episode is over.

        Returns:
            A list of applicable actions.
        """

        if not (self.terminated or self.truncated):
            return list(range(9))
        else:
            return []

    def get_observation_window_coordinates(self) -> tuple[int, int, int, int]:
        """Returns the top left and bottom right corner of the observation window.

        Returns:
            A tuple (top_left_x, top_left_y, bottom_right_x, bottom_right_y).
        """

        if not self.use_sliding_observation_window:
            # after the last step the agent could be outside the map
            position_inside_map_x = min(max(0, self.position[0]), self.map.width - 1)
            position_inside_map_y = min(max(0, self.position[1]), self.map.height - 1)

            tile_x = int(position_inside_map_x / TILE_WIDTH)
            tile_y = int(position_inside_map_y / TILE_HEIGHT)

            return (
                tile_x * TILE_WIDTH,
                tile_y * TILE_HEIGHT,
                tile_x * TILE_WIDTH + TILE_WIDTH - 1,
                tile_y * TILE_HEIGHT + TILE_HEIGHT - 1,
            )
        else:
            return (
                self.position[0] - self.sliding_observation_window_size,
                self.position[1] - self.sliding_observation_window_size,
                self.position[0] + self.sliding_observation_window_size,
                self.position[1] + self.sliding_observation_window_size,
            )



import copy
from pathlib import Path

import matplotlib as mpl
from matplotlib import colors
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PIL import Image, ImageDraw, ImageFont

font = ImageFont.load_default()
path_to_dir = Path(__file__).parent.resolve()
size = 100

white = colors.to_hex("w")
black = colors.to_hex("k")
red = colors.to_hex("r")
cyan = colors.to_hex("c")
blue = colors.to_hex("b")
magenta = colors.to_hex("m")
yellow = colors.to_hex("y")
grey = (196, 196, 196)
green = colors.to_hex("g")
start_color = (191, 128, 255)
neon_green = (128, 255, 0)
dark_green = (64, 127, 0)
light_grey = (220, 220, 220)
ice_color = (102, 212, 237)
sand_color = (234, 182, 118)
road_break_color = (44, 48, 48)
other_car_color = (0, 102, 0)
orange = (168, 62, 50)

line_colors = [red, yellow, cyan, blue, green, black]

marking_image = Image.new("RGBA", (size, size), white)
draw = ImageDraw.Draw(marking_image)
for i1 in range(11):
    for i2 in range(11):
        draw.ellipse((i1 * 10 - 4, i2 * 10 - 4, (i1 * 10 + 4, i2 * 10 + 4)), orange)
# marking_image.save("mark_test.png")


def get_tile(x, y, map, hide_start_line, marking=False, potentials=None):
    rectangle_color = (0, 0, 0) if not marking else orange
    rectangle_width = 2 if not marking else 5
    if map.feature_at(x, y, "wall"):
        wall = Image.new("RGBA", (size, size), grey)
        draw = ImageDraw.Draw(wall)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        draw.line((10, 10, size - 10, size - 10), fill=(0, 0, 0), width=2)
        draw.line((size - 10, 10, 10, size - 10), fill=(0, 0, 0), width=2)
        return wall
    elif map.feature_at(x, y, "start") and not hide_start_line:
        start = Image.new("RGBA", (size, size), start_color)
        draw = ImageDraw.Draw(start)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return start
    elif map.feature_at(x, y, "subgoal"):
        subgoal = Image.new("RGBA", (size, size), neon_green)
        draw = ImageDraw.Draw(subgoal)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return subgoal
    elif map.feature_at(x, y, "used subgoal"):
        used_subgoal = Image.new("RGBA", (size, size), dark_green)
        draw = ImageDraw.Draw(used_subgoal)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return used_subgoal
    elif map.feature_at(x, y, "final goal"):
        finish = Image.new("RGBA", (size, size), red)
        draw = ImageDraw.Draw(finish)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return finish
    elif map.feature_at(x, y, "ice"):
        ice = Image.new("RGBA", (size, size), ice_color)
        draw = ImageDraw.Draw(ice)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return ice
    # sAnd
    elif map.feature_at(x, y, "sand"):
        sand = Image.new("RGBA", (size, size), sand_color)
        draw = ImageDraw.Draw(sand)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return sand
    elif map.feature_at(x, y, "broken road"):
        road_break = Image.open(Path.joinpath(path_to_dir, "pics", "road_break.png"))
        draw = ImageDraw.Draw(road_break)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return road_break
    elif map.feature_at(x, y, "car_spawner"):
        beginning = Image.open(Path.joinpath(path_to_dir, "pics", "beginning.png"))
        draw = ImageDraw.Draw(beginning)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return beginning
    else:
        empty = Image.new("RGBA", (size, size), white)
        draw = ImageDraw.Draw(empty)
        draw.rectangle(
            ((0, 0), (size - 1, size - 1)),
            outline=rectangle_color,
            width=rectangle_width,
        )
        return empty


# specified font size
# font = ImageFont.truetype('SpaceMonoBoldItalic.ttf', 40)


def _transform_coordinates(x, y, offset=0.5, size=100):
    return (x + offset) * size, (y + offset) * size


def create_map(
    env,
    show_path=False,
    hide_start_line=False,
    show_observation_window=False,
):
    """create the map

    Args:
        env (Environment): Instance of rt game.
        show_path (bool, optional): show the taken path in the current episode. Defaults to False.
        hide_start_line (bool, optional): Hide the start line. Defaults to False.

    Returns:
        PIL: graphical way to represent the current state of game
    """
    h = env.map.width
    w = env.map.height
    result = Image.new(
        "RGBA",
        (size * h, size * w),
        (
            0,
            0,
            0,
        ),
    )

    for x in range(env.map.width):
        for y in range(env.map.height):
            tile = get_tile(x, y, env.map, hide_start_line)
            result.paste(
                tile,
                (x * size, y * size),
                mask=tile,
            )

    path = copy.copy(env.positions_path)
    path.append(path[-1])
    tile_path = copy.copy(env.tile_path)
    noise_path = copy.copy(env.noise_path)
    color = line_colors[0]
    o = 0.5
    circle_radius = 15
    rectangle_half_width = 50
    triangle_offset = 40
    car_half_width = 20
    draw = ImageDraw.Draw(result)

    if show_path:
        for i in range(len(path) - 1):
            f = path[i]
            t = path[i + 1]

            draw = ImageDraw.Draw(result)
            x1 = (f[0] + o) * size
            y1 = (f[1] + o) * size
            x2 = (t[0] + o) * size
            y2 = (t[1] + o) * size
            cx1 = x1 - 15
            cx2 = x1 + 15
            cy1 = y1 - 15
            cy2 = y1 + 15
            draw.line((x1, y1, x2, y2), fill=color, width=5)
            draw.ellipse((cx1, cy1, cx2, cy2), color)

        for x1, y1 in tile_path:
            x1, y1 = _transform_coordinates(x1, y1)
            draw.rectangle(
                (
                    (x1 - rectangle_half_width, y1 - rectangle_half_width),
                    (x1 + rectangle_half_width, y1 + rectangle_half_width),
                ),
                outline=color,
                width=5,
            )

        for x1, y1 in noise_path:
            x1, y1 = _transform_coordinates(x1, y1)
            draw.line(
                (
                    (x1 + triangle_offset, y1 + triangle_offset),
                    (x1, y1 - triangle_offset),
                    (x1 - triangle_offset, y1 + triangle_offset),
                    (x1 + triangle_offset, y1 + triangle_offset),
                ),
                fill=color,
                width=5,
            )

    new_position = copy.copy(env.position) + copy.copy(env.velocity)
    x0, y0 = copy.copy(env.position)
    x0, y0 = _transform_coordinates(x0, y0)
    x1, y1 = new_position
    x1, y1 = _transform_coordinates(x1, y1)

    draw.rectangle(
        (
            (x1 - rectangle_half_width, y1 - rectangle_half_width),
            (x1 + rectangle_half_width, y1 + rectangle_half_width),
        ),
        outline=(0, 0, 0),
        width=5,
    )
    if x0 < x1:
        draw.line((x0, y0, x1 - size * 0.5, y1 - size * 0.5), fill=(0, 0, 0), width=3)
        draw.line((x0, y0, x1 - size * 0.5, y1 + size * 0.5), fill=(0, 0, 0), width=3)
    elif x0 > x1:
        draw.line((x0, y0, x1 + size * 0.5, y1 - size * 0.5), fill=(0, 0, 0), width=3)
        draw.line((x0, y0, x1 + size * 0.5, y1 + size * 0.5), fill=(0, 0, 0), width=3)

    for car in [car["position"] for car in env.cars]:
        x, y = car
        x, y = _transform_coordinates(x, y)
        draw.rectangle(
            (
                x - car_half_width,
                y - car_half_width,
                x + car_half_width,
                y + car_half_width,
            ),
            other_car_color,
        )

    if show_observation_window:
        observation_window_coordinates = env.get_observation_window_coordinates()
        scaled_observation_window_coordinates = (
            observation_window_coordinates[0] * size,
            observation_window_coordinates[1] * size,
            (observation_window_coordinates[2] + 1) * size,
            (observation_window_coordinates[3] + 1) * size,
        )

        observation_window_mask = Image.new("RGBA", (size * h, size * w), (0, 0, 0, 99))
        ImageDraw.Draw(observation_window_mask).rectangle(
            scaled_observation_window_coordinates, fill=(0, 0, 0, 0)
        )

        result.paste(observation_window_mask, (0, 0), mask=observation_window_mask)

    return result


def print_heatmap(
    values,
    bounds=None,
    colormap=None,
    print_path=None,
    show=True,
    fig_size=None,
    font_size=None,
):
    """Builds a heatmap of a two dimensional array.

    :param values: 2D array from which the heatmap is generated
    :param bounds: bound array for the colors used in the heatmap
    :param colormap: a matplotlib colormap
    :param print_path: the path to where a png representation is saved
    :param show: boolean flag if the heatmap should be printed
    :returns: the figure object
    """
    if bounds is None:
        bounds = [-1, 0, 0.25, 0.5, 0.75, 0.9, 0.97, 0.99, 0.998, 1]
    if colormap is None:
        colormap = mpl.colors.ListedColormap(
            [
                "grey",
                "black",
                "red",
                "orange",
                "yellow",
                "lime",
                "limegreen",
                "green",
                "darkgreen",
            ]
        )
    norm = mpl.colors.BoundaryNorm(bounds, colormap.N)
    if fig_size is None:
        fig = plt.figure()
    else:
        fig = plt.figure(figsize=fig_size)
    fig.add_subplot(111)
    im = plt.pcolormesh(
        values, edgecolors="lightgray", linewidth=0.005, cmap=colormap, norm=norm
    )
    ax = plt.gca()
    plt.xticks([])
    plt.yticks([])
    ax.invert_yaxis()
    ax.set_aspect("equal")
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax)

    if font_size is not None:
        cbar.ax.tick_params(labelsize=font_size)

    if print_path is not None:
        plt.savefig(print_path)
    if show:
        plt.show()
    return fig
